#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2021  cutiful (https://github.com/cutiful)

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import os
import json
import http.server
import socketserver
from progress.bar import Bar
from urllib.parse import urlparse, parse_qs
from . import core

server_port = 13523
meow_origin = "https://purr.neocities.org"
meow_open_path = meow_origin + "/mastodon-archive-import/"

import_complete = False
bar = False

def meow(args):
    """
    Find and serve all archive files for Meow.
    """
    (username, domain) = args.user.split("@")

    status_file = domain + ".user." + username + ".json"
    data = core.load(status_file, required=True, quiet=True, combine=args.combine)

    media_dir = domain + ".user." + username
    media_files = []

    for collection in ["statuses", "favourites"]:
        for status in data[collection]:
            attachments = status["media_attachments"]
            if status["reblog"] is not None:
                attachments = status["reblog"]["media_attachments"]
            for attachment in attachments:
                if attachment["url"]:
                    path = urlparse(attachment["url"]).path
                    if path in media_files:
                        continue

                    # If we have it locally, set it to a relative path so Meow
                    # known to look in its local database. Otherwise, it'll
                    # still try to load the remote image.

                    file_name = media_dir + path
                    if os.path.isfile(file_name):
                        attachment["url"] = path
                        media_files.append(path)

    data["files"] = media_files

    global bar
    if len(media_files) > 0:
        bar = Bar("Exporting files", max = len(media_files) + 1)

    class Handler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            global import_complete
            global bar

            query = parse_qs(urlparse(self.path).query)

            if self.path == "/":
                self.send_response(200)
                self.send_header("Access-Control-Allow-Origin", meow_origin)
                self.send_header("Content-type", "application/json")
                self.end_headers()

                self.wfile.write(bytes(json.dumps(data), "utf-8"))

                if bar:
                    bar.next()
            elif "file" in query and query["file"][0] in media_files:
                self.send_response(200)
                self.send_header("Access-Control-Allow-Origin", meow_origin)
                self.end_headers()

                file_name = media_dir + query["file"][0]
                with open(file_name, "rb") as file:
                    self.wfile.write(file.read())

                if bar:
                    bar.next()
            elif "complete" in query:
                self.send_response(200)
                self.end_headers()

                import_complete = True

                if bar:
                    bar.finish()
            else:
                self.send_error(404)

        def log_message(self, format, *args):
            return

    def not_completed():
        global import_complete
        return not import_complete

    socketserver.TCPServer.allow_reuse_address = True
    with socketserver.TCPServer(("127.0.0.1", server_port), Handler) as httpd:
        print("Please, open Meow at", meow_open_path, "to continue!")
        while not_completed():
            httpd.handle_request()

    print("Export finished!")



#!/usr/bin/env python3
# Copyright (C) 2017  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import textwrap
import unicodedata
from . import core

def boosts(statuses):
    """
    Count boosts in statuses
    """
    i = 0
    for item in statuses:
        if item["reblog"] is not None:
            i += 1
    return i

def media(statuses):
    """
    Count media attachments in statuses
    """
    i = 0
    for item in statuses:
        i += len(item["media_attachments"])
    return i


def tags(statuses, include_boosts):
    """
    Count all the hashtags in statuses
    """
    count = {}
    for item in statuses:
        if include_boosts and item["reblog"] is not None:
            item = item["reblog"]
        for name in [tag["name"] for tag in item["tags"]]:
            if name in count:
                count[name] += 1
            else:
                count[name] = 1
    return count

def print_tags(statuses, max, include_boosts):
    """
    Print hashtags used in statuses
    """
    if max == -1:
        print("All the hashtags:")
    else:
        print("Top " + str(max) + " hashtags:")
    count = tags(statuses, include_boosts)
    most = sorted(count.keys(), key = lambda tag: -count[tag])
    print(textwrap.fill(" ".join(
        ["#"+tag+"("+str(count[tag])+")" for tag in most[0:max]])))

def emoji(statuses):
    """
    Count all the emoji in statuses
    """
    count = {}
    for item in statuses:
        for char in item["content"]:
            if unicodedata.category(char) == 'So':
                if char in count:
                    count[char] += 1
                else:
                    count[char] = 1
    return count

def print_emoji(statuses, min = 10, max_num = 30):
    """
    Print emoji used in statuses, sorted by frequency
    """
    print("Most frequeny Emoji:")
    count = emoji(statuses)
    count = {k: v for k, v in count.items() if v >= min }
    most = sorted(count.keys(), key = lambda emoji: -count[emoji])
    print(textwrap.fill(" ".join([emoji for emoji in most[0:max_num]])))

def report(args):
    """
    Report on your toots and favourites
    """

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required=True, quiet=True, combine=args.combine)

    if args.all:
        print("Considering the entire archive")
        statuses = data["statuses"]
        favourites = data["favourites"]
    else:
        print("Considering the last "
              + str(args.weeks)
              + " weeks")
        statuses = core.keep(data["statuses"], args.weeks)
        favourites = core.keep(data["favourites"], args.weeks)

    if "statuses" in data:
        print("Statuses:".ljust(20), str(len(statuses)).rjust(6))
        print("Boosts:".ljust(20), str(boosts(statuses)).rjust(6))
        print("Media:".ljust(20), str(media(statuses)).rjust(6))

        print()
        print_tags(statuses, args.top, args.include_boosts)

        if args.with_emoji:
            print()
            print_emoji(statuses)

    if "statuses" in data and "favourites" in data:
        print()

    if "favourites" in data:
        print("Favourites:".ljust(20), str(len(favourites)).rjust(6))
        print("Boosts:".ljust(20), str(boosts(favourites)).rjust(6))
        print("Media:".ljust(20), str(media(favourites)).rjust(6))

        print()
        print_tags(favourites, args.top, args.include_boosts)



#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
from . import core

def print_whitelist(args):

    (username, domain) = core.parse(args.user)

    whitelist = core.whitelist(domain, username)

    for account in whitelist:
        print(account)



#!/usr/bin/env python3
# Copyright (C) 2017-2019  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import argparse
from . import archive
from . import replies
from . import text
from . import context
from . import html
from . import media
from . import split
from . import expire
from . import report
from . import followers
from . import following
from . import whitelist
from . import mutuals
from . import login
from . import fix
from . import meow

def main():
    parser = argparse.ArgumentParser(
        description="""Archive your toots and favourites,
        and work with them.""",
        epilog="""Once you have created archives in the current directory, you can
        use 'all' instead of your account and the commands will be run
        once for every archive in the directory.""")

    subparsers = parser.add_subparsers()


    parser_content = subparsers.add_parser(
        name='archive',
        help='archive your toots and favourites')
    parser_content.add_argument("--no-favourites", dest='skip_favourites',
                                action='store_const',
                                const=True, default=False,
                                help='skip download of favourites')
    parser_content.add_argument("--with-mentions", dest='with_mentions',
                                action='store_const',
                                const=True, default=False,
                                help='download mentions (notifications where you are mentioned)')
    parser_content.add_argument("--with-followers", dest='with_followers',
                                action='store_const',
                                const=True, default=False,
                                help='download followers (people following you)')
    parser_content.add_argument("--with-following", dest='with_following',
                                action='store_const',
                                const=True, default=False,
                                help='download following (the people you follow)')
    parser_content.add_argument("--no-stopping", dest='stopping',
                                action='store_const',
                                const=False, default=True,
                                help='do not stop after seeing 10 duplicates')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=archive.archive)


    parser_content = subparsers.add_parser(
        name='replies',
        help='archive missing toots you replied to')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=replies.replies)


    parser_content = subparsers.add_parser(
        name='media',
        help='download media referred to by toots in your archive')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.add_argument("--collection", dest='collection',
                                choices=['statuses', 'favourites'],
                                default='statuses',
                                help='export statuses or favourites')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.set_defaults(command=media.media)


    parser_content = subparsers.add_parser(
        name='text',
        help='search and export toots in the archive as plain text')
    parser_content.add_argument("--reverse", dest='reverse', action='store_const',
                                const=True, default=False,
                                help='reverse output, oldest first')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.add_argument("--collection", dest='collection',
                                choices=['statuses', 'favourites', 'mentions', 'all'],
                                default='statuses',
                                help='export statuses, favourites, or mentions')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.add_argument("pattern", nargs='*',
                                help='regular expressions used to filter output')
    parser_content.set_defaults(command=text.text)


    parser_content = subparsers.add_parser(
        name='context',
        help='show a toot in context (i.e. with its ancestors and its descendants')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.add_argument("url",
                                help='URL of the toot to be included')
    parser_content.set_defaults(command=context.context)


    parser_content = subparsers.add_parser(
        name='html',
        help='export toots and media in the archive as static HTML')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.add_argument("--collection", dest='collection',
                                choices=['statuses', 'favourites'],
                                default='statuses',
                                help='export statuses or favourites')
    parser_content.add_argument("--toots-per-page", dest='toots',
                                metavar='N', type=int, default=2000,
                                help='how many toots per HTML page')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=html.html)

    parser_content = subparsers.add_parser(
        name='split',
        help='split an archive into two')
    parser_content.add_argument("--older-than", dest='weeks',
                                metavar='N', type=float, default=4,
                                help='split anything older than this many weeks')
    parser_content.add_argument("--confirmed", dest='confirmed',
                                action='store_const', const=True, default=False,
                                help='save the data after splitting')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=split.split)


    parser_content = subparsers.add_parser(
        name='expire',
        help='''delete older toots from the server and unfavour favourites
        if and only if they are in your archive''',
        epilog='''There is one problem you need to be aware of: if you expiring
        mentions, then the tool goes through all your notifications
        and looks at those of the type mention, and expires them if
        they are old enough. There are other types of notifications,
        however: follow, favourite, and reblog (at the time of this
        writing). As these are not archived, we also don't expire
        them. Thus, the list of notifications to look through keeps
        growing unless you use the "Clear notifications" menu in the
        Mastodon web client. Alternatively, you can use the
        --delete-other-notifications option together with
        --collection mentions.''')

    parser_content.add_argument("--collection", dest='collection',
                                choices=['statuses', 'favourites', 'mentions'],
                                default='statuses',
                                help='delete statuses, unfavour favourites, or clear mention notifications')
    parser_content.add_argument("--older-than", dest='weeks',
                                metavar='N', type=float, default=4,
                                help='expire anything older than this many weeks')
    parser_content.add_argument("--delete-other-notifications", dest='delete_others',
                                action='store_const', const=True, default=False,
                                help='clear follow, favourite, and reblog notifications')
    parser_content.add_argument("--confirmed", dest='confirmed',
                                action='store_const', const=True, default=False,
                                help='perform the expiration on the server')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=expire.expire)



    parser_content = subparsers.add_parser(
        name='report',
        help='''report some numbers about your toots and favourites''')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.add_argument("--all", dest='all', action='store_const',
                                const=True, default=False,
                                help='consider all toots (ignore --newer-than)')
    parser_content.add_argument("--newer-than", dest='weeks',
                                metavar='N', type=int, default=12,
                                help='only consider toots newer than this many weeks')
    parser_content.add_argument("--top", dest='top',
                                metavar='N', type=int, default=10,
                                help='only print the top N tags')
    parser_content.add_argument("--include-boosts", dest='include_boosts', action='store_const',
                                const=True, default=False,
                                help='include boosts')
    parser_content.add_argument("--with-emoji", dest='with_emoji', action='store_const',
                                const=True, default=False,
                                help='include emoji count')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=report.report)


    parser_content = subparsers.add_parser(
        name='followers',
        help='''find followers who never mention you''')
    parser_content.add_argument("--block", dest='block', action='store_const',
                                const=True, default=False,
                                help='...and block them')
    parser_content.add_argument("--all", dest='all', action='store_const',
                                const=True, default=False,
                                help='consider all toots (ignore --newer-than)')
    parser_content.add_argument("--newer-than", dest='weeks',
                                metavar='N', type=int, default=12,
                                help='require interaction within this many weeks (default is 12)')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=followers.followers)


    parser_content = subparsers.add_parser(
        name='following',
        help='''find people you are following but who never mention you''')
    parser_content.add_argument("--unfollow", dest='unfollow', action='store_const',
                                const=True, default=False,
                                help='...and unfollow them')
    parser_content.add_argument("--all", dest='all', action='store_const',
                                const=True, default=False,
                                help='consider all toots (ignore --newer-than)')
    parser_content.add_argument("--newer-than", dest='weeks',
                                metavar='N', type=int, default=12,
                                help='require mention within this many weeks (default is 12)')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=following.following)


    parser_content = subparsers.add_parser(
        name='mutuals',
        help='''find people you are following and who follow you back''')
    parser_content.add_argument("--pace", dest='pace', action='store_const',
                                const=True, default=False,
                                help='avoid timeouts and pace requests')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=mutuals.mutuals)


    parser_content = subparsers.add_parser(
        name='whitelist',
        help='''print the whitelist to help you debug problems''')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=whitelist.print_whitelist)

    parser_content = subparsers.add_parser(
        name='fix-boosts',
        help='''mark all the boosts as not deleted (triggering their deletion)''')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.add_argument("--confirmed", dest='confirmed',
                                action='store_const', const=True, default=False,
                                help='perform the change on the archive')
    parser_content.set_defaults(command=fix.fix_boosts)

    parser_content = subparsers.add_parser(
        name='login',
        help='login to the instance for testing purposes')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.set_defaults(command=login.login)

    parser_content = subparsers.add_parser(
        name='meow',
        help='import your backup into Meow, a browser-based export viewer (see https://purr.neocities.org/about/)')
    parser_content.add_argument("user",
                                help='your account, e.g. kensanata@octogon.social')
    parser_content.add_argument("--combine",
                                action="store_true",
                                help="combine archives in case they are split")
    parser_content.set_defaults(command=meow.meow)


    args = parser.parse_args()

    try:
        if hasattr(args, "command"):
            if hasattr(args, "user") and args.user == 'all':
                for user in core.all_accounts():
                    print(user)
                    args.user = user
                    args.command(args)
            else:
                args.command(args)
        else:
            parser.print_help()
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (C) 2017-2019  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017       Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from . import core

def archive(args):
    """
    Archive your toots and favourites from your Mastodon account
    """

    skip_favourites = args.skip_favourites
    with_mentions = args.with_mentions
    with_followers = args.with_followers
    with_following = args.with_following
    stopping = args.stopping

    (username, domain) = core.parse(args.user)

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file)

    mastodon = core.login(args)

    print("Get user info")

    try:
        user = mastodon.account_verify_credentials()
    except Exception as e:
        if "access token was revoked" in str(e):
            core.deauthorize(args)
             # retry and exit without an error
            archive(args)
            sys.exit(0)
        elif "Name or service not known" in str(e):
            print("Error: the instance name is either misspelled or offline",
              file=sys.stderr)
        else:
            print(e, file=sys.stderr)
        # exit in either case
        sys.exit(1)

    def complete(statuses, page, func = None):
        """
        Why aren't we using Mastodon.fetch_remaining(first_page)? It
        requires some metadata for the next request to be known. This
        is what the documentation says about
        Mastodon.fetch_next(previous_page): "Pass in the previous page
        in its entirety, or the pagination information dict returned
        as a part of that pages last status (‘_pagination_next’)."
        When we last updated our archive, however, there was no next
        page: we got all the pages there were. So the archive will
        have a ton of _pagination_prev keys but no _pagination_next
        keys. That's why we fetch it all over again. Expiry helps,
        obviously.
        """
        seen = { str(status["id"]): status for status in statuses }
        progress = core.progress_bar()

        # define function such that we can return from the inner and
        # from the outer loop
        def process(page):
            count = 0
            duplicates = 0
            while len(page) > 0:
                progress()
                for item in page:
                    status = item
                    # possibly a notification containing a status
                    if "status" in item:
                        status = item["status"]
                    if status and "id" in status:
                        id = str(status["id"])
                        if not id in seen:
                            if func is None or func(item):
                                seen["id"] = status
                                statuses.append(status)
                                count = count + 1
                        else:
                            duplicates = duplicates + 1
                            if duplicates > 10 and stopping:
                                print() # at the end of the progress bar
                                print("Seen 10 duplicates, stopping now.")
                                print("Use --no-stopping to prevent this.")
                                return count
                page = mastodon.fetch_next(page)
                if page is None:
                    print() # at the end of the progress bar
                    return count
            # if len(page) was 0
            return count

        count = process(page)
        print("Added a total of %d new items" % count)
        return statuses

    def keep_mentions(notifications):
        return [x.status for x in notifications if x.type == "mention"]

    if data is None or not "statuses" in data or len(data["statuses"]) == 0:
        print("Get all statuses (this may take a while)")
        statuses = mastodon.account_statuses(user["id"], limit=100)
        statuses = mastodon.fetch_remaining(
            first_page = statuses)
    else:
        print("Get new statuses")
        statuses = complete(data["statuses"], mastodon.account_statuses(user["id"], limit=100))

    if skip_favourites:
        print("Skipping favourites")
        if data is None or not "favourites" in data:
            favourites = []
        else:
            favourites = data["favourites"]
    elif data is None or not "favourites" in data or len(data["favourites"]) == 0:
        print("Get favourites (this may take a while)")
        favourites = mastodon.favourites()
        favourites = mastodon.fetch_remaining(
            first_page = favourites)
    else:
        print("Get new favourites")
        favourites = complete(data["favourites"], mastodon.favourites())

    if not with_mentions:
        print("Skipping mentions")
        if data is None or not "mentions" in data:
            mentions = []
        else:
            mentions = data["mentions"]
    elif data is None or not "mentions" in data or len(data["mentions"]) == 0:
        print("Get notifications and look for mentions (this may take a while)")
        notifications = mastodon.notifications(limit=100)
        notifications = mastodon.fetch_remaining(
            first_page = notifications)
        mentions = keep_mentions(notifications)
    else:
        print("Get new notifications and look for mentions")
        is_mention = lambda x: "type" in x and x["type"] == "mention"
        mentions = complete(data["mentions"], mastodon.notifications(limit=100), is_mention)

    if not with_followers:
        print("Skipping followers")
        if data is None or not "followers" in data:
            followers = []
        else:
            followers = data["followers"]
    else:
        print("Get followers (this may take a while)")
        followers = mastodon.account_followers(user.id, limit=100)
        followers = mastodon.fetch_remaining(
            first_page = followers)

    if not with_following:
        print("Skipping following")
        if data is None or not "following" in data:
            following = []
        else:
            following = data["following"]
    else:
        print("Get following (this may take a while)")
        following = mastodon.account_following(user.id, limit=100)
        following = mastodon.fetch_remaining(
            first_page = following)

    data = {
        'account': user,
        'statuses': statuses,
        'favourites': favourites,
        'mentions': mentions,
        'followers': followers,
        'following': following,
    }

    print("Saving %d statuses, %d favourites, %d mentions, %d followers, and %d following" % (
        len(statuses),
        len(favourites),
        len(mentions),
        len(followers),
        len(following)))

    core.save(status_file, data)



#!/usr/bin/env python3
# Copyright (C) 2017–2019  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017       Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import json
import time
import urllib.request
from urllib.error import HTTPError
from urllib.error import URLError
from progress.bar import Bar
from urllib.parse import urlparse
from . import core

def media(args):
    """
    Download all the media files linked to from your archive
    """

    pace = hasattr(args, 'pace') and args.pace

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    media_dir = domain + '.user.' + username
    data = core.load(status_file, required=True, quiet=True, combine=args.combine)

    urls = []
    preview_urls_count=0
    for status in data[args.collection]:
        attachments = status["media_attachments"]
        if status["reblog"] is not None:
            attachments = status["reblog"]["media_attachments"]
        for attachment in attachments:
                if attachment["preview_url"]:
                        urls.append(attachment["preview_url"])
                        preview_urls_count += 1
                if attachment["url"]:
                        urls.append(attachment["url"])

    print("%d urls in your backup (%d are previews)" % (len(urls), preview_urls_count))

    bar = Bar('Downloading', max = len(urls))

    errors = 0

    # start downloading the missing files from the back
    for url in reversed(urls):
        bar.next()
        path = urlparse(url).path
        file_name = media_dir + path
        if not os.path.isfile(file_name):
            dir_name =  os.path.dirname(file_name)
            os.makedirs(dir_name, exist_ok = True)
            try:
                req = urllib.request.Request(
                    url, data=None,
                    headers={'User-Agent': 'Mastodon-Archive/1.3 '
                             '(+https://github.com/kensanata/mastodon-backup#mastodon-archive)'})
                try:
                  with urllib.request.urlopen(req) as response, open(file_name, 'wb') as fp:
                    data = response.read()
                    fp.write(data)
                except HTTPError as he:
                  print("\nFailed to open " + url + " during a media request.")
                except URLError as ue:
                  print("\nFailed to open " + url + " during a media request.")
            except OSError as e:
                print("\n" + e.msg + ": " + url, file=sys.stderr)
                errors += 1
            if pace:
                time.sleep(1)

    bar.finish()

    if errors > 0:
        print("%d downloads failed" % errors)



#!/usr/bin/env python3
# Copyright (C) 2019  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from progress.bar import Bar
from . import core

def replies(args):
    """
    Archive the statuses you replied to
    """

    (username, domain) = core.parse(args.user)

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required = True)

    mastodon = core.login(args)

    print("Get user info")

    try:
        user = mastodon.account_verify_credentials()
    except Exception as e:
        if "access token was revoked" in str(e):
            core.deauthorize(args)
             # retry and exit without an error
            archive(args)
            sys.exit(0)
        elif "Name or service not known" in str(e):
            print("Error: the instance name is either misspelled or offline",
              file=sys.stderr)
        else:
            print(e, file=sys.stderr)
        # exit in either case
        sys.exit(1)

    index = {} # mapping ids to statuses
    missing = [] # ids we need to fetch

    for collection in ["statuses",
                       "favourites",
                       "mentions",
                       "replies"]:
        if collection not in data:
            print("No %s in this archive..." % collection)
        else:
            statuses = data[collection];
            print("Indexing %d %s..." % (len(statuses), collection))
            for status in statuses:

                if status["reblog"] is not None:
                    status = status["reblog"]

                # only accept one status per id
                if status["id"] in index:
                    pass
                else:
                    index[status["id"]] = 1
    print("Indexed %d statuses..." % (len(index)))

    print("Counting missing replies...")
    for status in data["statuses"]:
        # skip boosts
        if status["reblog"] is None and status["in_reply_to_id"] is not None:
            if status["in_reply_to_id"] not in index:
                missing.append(status["in_reply_to_id"])
    print("Missing %d originals..." % (len(missing)))

    if len(missing) > 300:
        print("Given the typical rate limit of 300 requests per 5 minutes, "
              "this will take about %d minutes" % (len(missing) // 300 * 5))

    if len(missing) > 0:
        if not "replies" in data:
            replies = []
        else:
            replies = data["replies"]

        bar = Bar('Fetching', max = len(missing))

        for id in missing:
            try:
                status = mastodon.status(id)
                replies.append(status)
            except Exception as e:
                if  "not found" in str(e):
                    pass
                else:
                    print(e, file=sys.stderr)

            bar.next()

        bar.finish()

        data["replies"] = replies
        core.save(status_file, data)



#!/usr/bin/env python3
# Copyright (C) 2017-2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from mastodon import Mastodon
import sys
import os.path
import datetime
import json
import glob
import re
import shutil

def progress_bar(chars="▏▎▍▌▋▊▉█"):
    """
    Return a progress bar updater which you can then call.
    """
    n = 0
    def progress():
        nonlocal n
        if n != 0:
            sys.stdout.write("\b")
        sys.stdout.write(chars[n])
        sys.stdout.flush()
        n = (n + 1) % len(chars)

    return progress

def parse(account):
    """
    Parse account into username and domain.
    """
    try:
        (username, domain) = account.split('@')
        return username, domain;
    except ValueError:
        print("The account has to have the form username@domain", file=sys.stderr)
        sys.exit(3)

def read(args):
    """
    Login to your Mastodon account with read-only permissions.
    """
    return login(args)

def readwrite(args):
    """
    Login to your Mastodon account with read and write permissions.
    Note that you will get an error when your authorization token is a
    read-only token. If this happens, you need to deauthorize and try
    again.
    """
    try:
        # this is what we expect
        return login(args, scopes = ['read', 'write'])
    except Exception as e:
        # on some instances, there's this problem with getting a
        # bigger scope than requested, so just do it again and hope
        # for the best... (dealing with MastodonAPIError: Granted
        # scopes "follow read write" differ from requested scopes
        # "read write".)
        return login(args, scopes = ['follow', 'read', 'write'])

def readwritefollow(args):
    """
    Login to your Mastodon account with read, write and follow permissions.
    Note that you will get an error when your authorization token is a
    read-only token. If this happens, you need to deauthorize and try
    again.
    """
    return login(args, scopes = ['follow', 'read', 'write'])

def deauthorize(args):
    """
    Deauthorize the account.
    """
    app = App(args.user)
    app.deauthorize()

def login(args, scopes=('read',)):
    """
    Login to your Mastodon account
    """
    pace = hasattr(args, 'pace') and args.pace
    app = App(args.user, scopes)
    return app.login(pace)


class App:
    """
    Client application to register, authorize and login with your Mastodon
    account.
    """

    def __init__(self, user, scopes=('read',), name="mastodon-archive"):

        self.username, self.domain = user.split("@")
        self.url = "https://" + self.domain
        self.name = name
        self.scopes = scopes
        self.client_secret = self.domain + ".client.secret"
        self.user_secret = self.domain + ".user." + self.username + ".secret"

    def register(self):
        """
        Register application and saves client secret.
        """
        print("Registering app")
        Mastodon.create_app(
            self.name,
            api_base_url=self.url,
            scopes=self.scopes,
            to_file=self.client_secret
        )

    def authorize(self):
        """
        Tries to authorize via OAuth API, and save access token. If it fails
        fallsback to username and password.
        """
        url = self.url
        client_secret = self.client_secret
        user_secret = self.user_secret
        scopes = self.scopes
        print("This app needs access to your Mastodon account.")

        mastodon = Mastodon(client_id=client_secret, api_base_url=url)

        url = mastodon.auth_request_url(client_id=client_secret, scopes=scopes)

        print("Visit the following URL and authorize the app:")
        print(url)

        print("Then paste the access token here:")
        token = sys.stdin.readline().rstrip()

        try:
            # on the very first login, --pace has no effect
            mastodon.log_in(code=token, to_file=user_secret, scopes=scopes)

        except Exception:

            print("Sadly, that did not work. On some sites, this login mechanism")
            print("(namely OAuth) seems to be broken. There is an alternative")
            print("if you are willing to trust us with your password just this")
            print("once. We need it just this once in order to get an access")
            print("token. We won't save it. If you don't want to try this, use")
            print("Ctrl+C to quit. If you want to try it, please provide your")
            print("login details.")

            sys.stdout.write("Email: ")
            sys.stdout.flush()
            email = sys.stdin.readline().rstrip()
            sys.stdout.write("Password: ")
            sys.stdout.flush()
            password = sys.stdin.readline().rstrip()

            # on the very first login, --pace has no effect
            mastodon.log_in(
                username=email,
                password=password,
                to_file=user_secret,
                scopes=scopes
            )

        return mastodon

    def deauthorize(self):
        """
        Deauthorize by deleting the file containing the authorization token.
        """
        user_secret = self.user_secret
        client_secret = self.client_secret
        if os.path.isfile(user_secret):
            os.remove(user_secret)
        if os.path.isfile(client_secret):
            os.remove(client_secret)

    def login(self, pace=False):
        """
        Register app, authorize and return an instance of ``Mastodon``
        """
        url = self.url
        client_secret = self.client_secret
        user_secret = self.user_secret

        if not os.path.isfile(client_secret):
            self.register()

        if not os.path.isfile(user_secret):
            mastodon = self.authorize()

        else:

            if pace:

                # in case the user kept running into a General API problem
                mastodon = Mastodon(
                    client_id=client_secret,
                    access_token=user_secret,
                    api_base_url=url,
                    ratelimit_method="pace",
                    ratelimit_pacefactor=0.9,
                    request_timeout=300
                )

            else:

                # the defaults are ratelimit_method='wait',
                # ratelimit_pacefactor=1.1, request_timeout=300
                mastodon = Mastodon(
                    client_id=client_secret,
                    access_token=user_secret,
                    api_base_url=url
                )

        return mastodon

def load(file_name, required=False, quiet=False, combine=False):
    """
    Load the JSON data from a file.
    """

    if required and not os.path.isfile(file_name):
        print("You need to create an archive, first", file=sys.stderr)
        sys.exit(2)

    if os.path.isfile(file_name) and os.path.getsize(file_name) > 0:

        def _json_load(fname):
            if not quiet:
                print("Loading existing archive:", fname)

            with open(fname, mode='r', encoding='utf-8') as fp:
                return json.load(fp)

        data = _json_load(file_name)
        if combine:
            # Load latest archive first to keep chronological order
            archives = list(
                reversed(glob.glob(file_name.replace(".json", ".*.json")))
            )

            if required and not quiet and not archives:
                print("Warning: No split archives to combine", file=sys.stderr)

            # Merge dictionaries loaded from JSON archives
            for archive in archives:
                archived_data = _json_load(archive)

                for collection in ["statuses", "favourites", "mentions"]:
                    data[collection].extend(archived_data[collection])

        return data

    return None

def save(file_name, data):
    """
    Save the JSON data in a file. If the file exists, rename it,
    just in case.
    """
    date_handler = lambda obj: (
        obj.isoformat()
        if isinstance(obj, (datetime.datetime, datetime.date))
        else None)

    if os.path.isfile(file_name):
        backup_file = file_name + '~'
        print("Backing up", file_name, "to", backup_file)
        if os.path.isfile(backup_file):
            ans = ""
            while ans.lower() not in ("y", "n", "yes", "no"):
                ans = input(
                    "Backup: {} exists! Overwrite (yes/no)? ".format(backup_file)
                )

            if ans.lower()[0] == "y":
                shutil.copy2(file_name, backup_file)
            else:
                print("Exiting to avoid overwriting backup.", file=sys.stderr)
                sys.exit(0)

    with open(file_name, mode = 'w', encoding = 'utf-8') as fp:
        data = json.dump(data, fp, indent = 2, default = date_handler)

def all_accounts():
    """
    Return all the known user accounts in the current directory.
    """
    archives = glob.glob('*.user.*.json');
    if not archives:
        print("You need to create an archive, first", file=sys.stderr)
        sys.exit(2)
    else:
        users = []
        for archive in archives:
            m = re.match(r"(.*)\.user\.(.*)\.json", archive)
            if m:
                users.append("%s@%s" % m.group(2, 1))
        return users

def keep(statuses, weeks):
    """
    Return all statuses newer than some weeks
    """

    delta = datetime.timedelta(weeks = weeks)
    cutoff = datetime.datetime.today() - delta

    def matches(status):
        created = datetime.datetime.strptime(status["created_at"][0:10], "%Y-%m-%d")
        return created >= cutoff

    return list(filter(matches, statuses))

def whitelist(domain, username):
    file_name = domain + '.user.' + username + '.whitelist.txt'
    whitelist = set()
    if os.path.isfile(file_name):
        with open(file_name, mode = 'r', encoding = 'utf-8') as fp:
            for line in fp:
                # kensanata
                # kensanata@dice.camp
                # Alex Schroeder <kensanata@dice.camp>
                m = re.search(r"([a-zA-Z0-9.-]+@[a-zA-Z0-9.-]+)", line)
                if not m:
                    m = re.search(r"([a-zA-Z0-9.-]+)", line)
                if m:
                    whitelist.add(m.group(1))
        print("%d accounts are on the whitelist" % len(whitelist))
    else:
        print("There is no whitelist")
    return whitelist



#!/usr/bin/env python3
# Copyright (C) 2017  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017  Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import html2text
import datetime
import dateutil.parser
from urllib.parse import urlparse
from . import core

header_template = '''\
<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8"/>
<title>%s</title>
<style type="text/css">
body {
        font-family: sans-serif;
        background: #191b22;
        font-size: 14px;
        line-height: 18px;
        font-weight: 400;
        color: #fff;
        text-rendering: optimizelegibility;
        -webkit-font-feature-settings: "kern";
        font-feature-settings: "kern";
        -webkit-text-size-adjust: none;
        -moz-text-size-adjust: none;
        -ms-text-size-adjust: none;
        text-size-adjust: none;
        -webkit-tap-highlight-color: transparent;
}
.column {
        background: #282c37;
        padding: 10px;
        margin: auto;
        max-width: 80ex;
}
h1 {
        color: #fff;
        font-size: 20px;
        line-height: 27px;
        font-weight: 500;
        overflow: hidden;
        text-overflow: ellipsis;
        text-align: center;
}
.account {
        text-align: center;
}
.account .nick {
        color: #2b90d9;
        margin-bottom: 10px;
}
.account .bio, .bio a, .bio a:visited {
        color: #d9e1e8;
}
.meta {
        color: #606984;
        font-size: 15px;
}
.meta strong {
        font-weight: 500;
}
a {
        text-decoration: none;
}
a:hover {
        text-decoration: underline;
}
.wrapper {
        padding-top: 10px;
        border-top: 1px solid #393f4f;
}
.boosted a {
        color: #606984;
}
.boosted a:visited {
        color: #606984;
}
.boosted {
        padding-bottom: 10px;
}
.meta a {
        color: #fff;
}
.meta a:visited {
        color: #fff;
}
.meta a.time {
        color: inherit;
}
.meta a.time:visited {
        color: inherit;
}
.meta .time {
        display: block;
        float: right;
}
.media {
        margin-top: 8px;
        margin-bottom: 8px;
        height: 110px;
        overflow: hidden;
}
.media a {
        display: block;
        float: left;
        position: relative;
        border: 1px solid #282c37;
        box-sizing: border-box;
        overflow: hidden;
}
.more a {
        max-width: 50%%;
        max-height: 50%%;
        display: block;
}
/* http://jonrohan.codes/fieldnotes/vertically-center-clipped-image/ */
.media img {
        position: relative;
        top: 50%%;
        transform: translateY(-50%%);
}
nav a, .content a {
        color: #d9e1e8;
}
nav a:visited, .content a:visited {
        color: #d9e1e8;
}
nav {
        padding: 10px 0;
        border-top: 1px solid #393f4f;
}
footer nav {
        padding-bottom: 0;
        border-top: 1px solid #393f4f;
}
.invisible {
        display: none;
}
</style>
</head>
<body>
<div class="column">
<header>
<h1 class="name">%s</h1>
<div class="account">
<p class="nick">@%s</p>
<div class="bio">
%s
</div>
</div>
</header>
%s\
'''

footer_template = '''\
<footer>
%s\
<footer>
</div>
</body>
</html>
'''

nav_template = '''\
<nav>
%s\
%s\
</nav>
'''

previous_template = '''\
<a class="previous" href="%s">Previous</a>
'''

next_template = '''\
<a class="next" href="%s">Next</a>
'''

boost_template = '''\
<div class="boosted">
<a class="name" href="%s">%s</a> boosted
</div>
'''

status_template = '''\
<div class="status">
<div class="meta">
<strong class="name"><a href="%s">%s</a></strong>
<span class="nick">@%s</span>
<a class="time" href="%s">%s</a>
</div>
<div class="content">
%s
</div>
</div>
'''

media_template = '''\
<div class="media %s">
%s</div>\
'''

preview_template = '''\
<a href="%s"><img src="%s"/></a>
'''

wrapper_template = '''\
<div class="wrapper">
%s%s%s
</div>
'''

def file_url(media_dir, url1, url2=None):
    for url in [url1, url2]:
        if url is not None:
            path = urlparse(url).path
            if os.path.isfile(media_dir + path):
                return media_dir + path
    return "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAADIAAAAyCAYAAAAeP4ixAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4QsUETQjvc7YnAAAACZpVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVAgb24gYSBNYWOV5F9bAAADfElEQVRo3u1ZTUhUURg997sPy9TnDxk4IeqkmKghCKLBLNq2MCXQTXsXomFMIdK0EIoIskWLcNVKaFEulBIzxEgQLBRFXZmiaIOkboSe2Lv3tsg3+Hr96Mx7OhNzl/PuvHfPPd/5zvfdy5719Cj8B4Pwn4wkkCSQRAGilIIQ4q9zpJSQUrr6Xc2tF0kpQURgjGF2fh4zc3P4vLyMza0t27xzubkoKy1FoLYW532+yP/iBohhGHjR349P09M/qSaCUs7M/nVzE1vb23g/Po6zOTkItrcjPS0NnPOYvs9i9RGlFJ739eHj1BSI6EghY81va2nBxZKSmJiJidPvpon+wUHMzs1FwgsAGGN/3rkDz6z5T3t7sR4O/1NbnoUWA3AqJQXm/gKsHS7Iz0egrg7lZWXI1PUI6PmFBbwcGMDW9rbjXQ8eP8aznp4T1AhjICIIIVBVWYkbzc1IPX0apmnaQkXjHJXl5ai6dAmvBgbwbmzMwdSHiQlcrqmJSi+uiD0jPR3BtjZkZ2VFwkXTNMdCrQVer6/H6toaFpeWbJqamJxEoK7u+DVCRKiqqMD9UCgSQocRrJQSTQ0NNhBKKSyvrJyM2IkIvry8QwM4yM55ny++nD2alPm3rHbihnjUKiC8seEAl52VlVhFIxFhcGjIxiZjDDXV1VF7ybEz8t008SUcxsy+iR5k6drVq78ta+KOEdM0sbe3h4dPnjieNTU2QggRtX6OjREhBKSUCN69C8ZYZOeJCBeKinAlEIh/sVs7fbOz0waCc46c7Gzcam2N/w7RAtF2546ttOecI1PX0d3VFbUujg2IlBKcc7QGg7Zql3MOPSMD90MhKKVc8RVPgRARbodC4JzbQJxJTcWDe/dcA+GpRqSUeD08jG+GYauphBB41N3tWovrOSNEhDcjI46OsbOjw1Hixy0jUkqsh8OOEkTXdRTk5yfOuZZSCqtraw4gZSUliXVAp5TCzs6Oo5bK1PWY+vIT0Yhpmo6MpGmaK54RN9Vv4gPxoKHyViMAjN1dmx6EEDAMA17dKrHkjVWcDU9LlF9dnYhcd3TPnX1xacl2AEdEKPb7Uez3ewLGUyBvR0cj58La/imjv7AwcYBYYEwhbJlLKpUUezJrudGPaAeuBzTOQR46u+YViGK/39anW78lVPq1Fu0vLExsH/F60cmslQSSBHL08QPK53LVsfanXQAAAABJRU5ErkJggg=="

def write_status(fp, media_dir, status):
    boost = "";
    if status["reblog"] is not None:
        user = status["account"]
        boost = boost_template % (
            user["url"],
            user["display_name"])
        # display the boosted status instead
        status = status["reblog"]

    user = status["account"]
    info = status_template % (
        user["url"],
        user["display_name"],
        user["username"],
        status["url"],
        dateutil.parser.parse(
            status["created_at"]).strftime(
                "%Y-%m-%d %H:%M"),
        status["content"])

    media = ''
    attachments = status["media_attachments"]

    if len(attachments) > 0:
        previews = []
        for attachment in attachments:
            previews.append(preview_template % (
                file_url(media_dir, attachment["url"]),
                file_url(media_dir, attachment["preview_url"])))

        media = media_template % (
                'more' if len(attachments) > 1 else 'one',
                ''.join(previews))

    html = wrapper_template % (boost, info, media)
    fp.write(html)

def html_file(domain, username, collection, page):
    return (domain + '.user.' + username + '.'
            + collection + '.' + str(page) + '.html')

def html(args):
    """
    Convert toots and media files to static HTML
    """

    toots_per_page = args.toots
    collection = args.collection
    combine = args.combine

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    media_dir = domain + '.user.' + username
    base_url = 'https://' + domain
    data = core.load(status_file, required=True, combine=combine)
    user = data["account"]
    statuses = data[collection]

    if len(statuses) > 0:

        (pages, offset) = divmod(len(statuses), toots_per_page)
        page = 0

        while (page <= pages):

            if pages == 0:
                nav_html = ""
            else:
                if (page == 0):
                    previous_html = ""
                else:
                    previous_html = previous_template % (
                        html_file(domain, username, collection, page - 1))

                if (page < pages):
                    next_html = next_template % (
                        html_file(domain, username, collection, page + 1))
                else:
                    next_html = ""

                nav_html = nav_template % (previous_html, next_html)

            file_name = html_file(domain, username, collection, page)

            with open(file_name, mode = 'w', encoding = 'utf-8') as fp:

                print("Writing %s" % file_name)

                html = header_template % (
                    user["display_name"],
                    user["display_name"],
                    user["username"],
                    user["note"],
                    nav_html)

                # This forces UTF-8 independent of terminal capabilities, thus
                # avoiding problems with LC_CTYPE=C and other such issues.
                # This works well when redirecting output to a file, which
                # will then be an UTF-8 encoded file.
                fp.write(html)

                # Assume 184 toots, 100 toots per page:
                # page 0 is 0:84, page 1 is 84:184
                for status in statuses[
                        max(0, toots_per_page * (page - 1) + offset)
                        : toots_per_page * page + offset]:
                    write_status(fp, media_dir, status)

                fp.write(footer_template % nav_html)
                page += 1
                toots = toots_per_page



#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from progress.bar import Bar
from datetime import timedelta, datetime
from . import core

def is_lurker(account, mentions):
    for mention in mentions:
        if account["id"] == mention["account"]["id"]:
            return False
    return True

def find_lurkers(followers, whitelist, mentions):
    return [x for x in followers
            if x["acct"] not in whitelist
            and is_lurker(x, mentions)]

def followers(args):
    """
    List followers who never mention you
    """

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required=True, quiet=True, combine=True)

    # Print both error messages if the data is missing
    error = 0
    if "mentions" not in data or len(data["mentions"]) == 0:
        print("You need to run 'mastodon-archive archive --with-mentions'",
              file=sys.stderr)
        error = 5
    if "followers" not in data or len(data["followers"]) == 0:
        print("You need to run 'mastodon-archive archive --with-followers'",
              file=sys.stderr)
        error = 6
    if error > 0:
        sys.exit(error)

    if args.all:
        print("Considering the entire archive")
        mentions = data["mentions"]
    else:
        print("Considering the last "
              + str(args.weeks)
              + " weeks")
        mentions = core.keep(data["mentions"], args.weeks)

    whitelist = core.whitelist(domain, username)

    if args.block:
        mastodon = core.readwrite(args)
        accounts = find_lurkers(data["followers"], whitelist, data["mentions"])

        bar = Bar('Blocking', max = len(accounts))

        for account in accounts:
            bar.next()
            try:
                mastodon.account_block(account["id"])
            except Exception as e:
                if "authorized scopes" in str(e):
                    print("\nWe need to authorize the app to make changes to your account.")
                    core.deauthorize(args)
                    mastodon = core.readwrite(args)
                    # retry
                    mastodon.account_block(account["id"])
                else:
                    print(e, file=sys.stderr)

        bar.finish()

    else:
        accounts = find_lurkers(data["followers"], whitelist, data["mentions"])
        for account in sorted(accounts, key=lambda account:
                              account["display_name"] or account["username"]):
            print("%s <%s>" % (account["display_name"] or account["username"],
                               account["acct"]))



#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
from . import core

def fix_boosts(args):
    """
    Go through all the boosts and mark them as undeleted.
    """

    confirmed = args.confirmed

    if not confirmed:

        print("This is a dry run and nothing will be changed.\n"
              "Instead, we'll just list what would have happened.\n"
              "Use --confirmed to actually do it.")

    (username, domain) = core.parse(args.user)

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required=True, combine=args.combine)
    n = 0

    for status in data["statuses"]:
        if status["reblog"] and "deleted" in status and status["deleted"]:
            del status["deleted"]
            n = n + 1
            if n == 1 and not confirmed:
                print("Some examples:")
            if n <= 20 and not confirmed:
                print(str(n) + " " + status["reblog"]["url"])

    if confirmed and n > 0:

        print("Saving updated data to", status_file)
        core.save(status_file, data)

    elif confirmed:

        print("No boosted statuses were undeleted")

    else:

        print("Would have marked " + str(n) + " statuses as not deleted")



#!/usr/bin/env python3
# Copyright (C) 2017  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017  Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import math
from progress.bar import Bar
from datetime import timedelta, datetime
from random import shuffle
import signal
import html2text
import textwrap
from . import core

h = html2text.HTML2Text()
h.ignore_links = True

def text(status):
    text = textwrap.fill(h.handle(status["content"])).lstrip();
    text = text.replace("\n", " ")
    if len(text) > 50:
        text = text[0:50] + '...'
    return "%s \"%s\"" % (status["created_at"][0:10], text)

def delete(mastodon, collection, status):
    """
    Delete toot, unfavour favourite, or dismiss notification and mark
    it as deleted. The "record not found" error is handled elsewhere.
    """
    if collection == 'statuses':
        if status["reblog"]:
            mastodon.status_unreblog(status["reblog"]["id"])
        else:
            mastodon.status_delete(status["id"])
    elif collection == 'favourites':
        mastodon.status_unfavourite(status["id"])
    elif collection == 'mentions':
        mastodon.notifications_dismiss(status["id"])
    status["deleted"] = True

def expire(args):
    """
    Expire toots: delete toots, unfavour favourites, or dismiss
    notifications older than a few weeks
    """

    confirmed = args.confirmed
    collection = args.collection
    delete_others = args.delete_others

    if (delete_others and collection != 'mentions'):
        print("The --delete-others option can only be used "
              "together with --collection mentions.",
              file=sys.stderr)
        sys.exit(4)

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required = True)

    if confirmed:
        mastodon = core.readwrite(args)
    else:
        print("This is a dry run and nothing will be expired.\n"
              "Instead, we'll just list what would have happened.\n"
              "Use --confirmed to actually do it.")

    delta = timedelta(weeks = args.weeks)
    cutoff = datetime.today() - delta

    def matches(status):
        created = datetime.strptime(status["created_at"][0:10], "%Y-%m-%d")
        deleted = "deleted" in status and status["deleted"] == True
        pinned = "pinned" in status and status["pinned"] == True
        return created < cutoff and not deleted and not pinned

    statuses = list(filter(matches, data[collection]))
    n_statuses = len(statuses)
    shuffle(statuses)

    if (n_statuses == 0):
        print("No " + collection + " are older than %d weeks" % args.weeks,
              file=sys.stderr)

    if confirmed and n_statuses > 0:

        bar = Bar('Expiring', max = len(statuses))
        error = ''

        def signal_handler(signal, frame):
            print("\nYou pressed Ctrl+C! Saving data before exiting!")
            core.save(status_file, data)
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        for status in statuses:
            try:
                delete(mastodon, collection, status)
            except Exception as e:
                if "authorized scopes" in str(e):
                    print("\nWe need to authorize the app to make changes to your account.")
                    core.deauthorize(args)
                    mastodon = core.readwrite(args)
                    # retry
                    delete(mastodon, collection, status)
                elif "not found" in str(e):
                    status["deleted"] = True
                elif "Name or service not known" in str(e):
                    error = "Error: the instance name is either misspelled or offline"
                else:
                    print(e, file=sys.stderr)
            bar.next()

        bar.finish()

        if error:
            print(error, file=sys.stderr)

        core.save(status_file, data)

    elif n_statuses > 0:

        for status in statuses:
            if collection == 'statuses':
                print("Delete: " + text(status))
            elif collection == 'favourites':
                print("Unfavour: " + text(status))
            elif collection == 'mentions':
                print("Dismiss: " + text(status))

    if delete_others:
        print('Dismissing other notifications')
        progress = core.progress_bar()

        # unlike above where we're getting the created_at value from
        # the JSON file where the date comes in iso format... see
        # core.save
        def others(notifications):
            return [x for x in notifications
                    if x.type != "mention"
                    and x["created_at"].replace(tzinfo = None) < cutoff]

        mastodon = core.login(args)
        notifications = mastodon.notifications(limit=100)
        total = 0
        dismissed = 0
        while (notifications):
            progress()
            total += len(notifications)
            for notification in others(notifications):
                if confirmed:
                    try:
                        mastodon.notifications_dismiss(notification["id"])
                        dismissed += 1
                    except Exception as e:
                        if "authorized scopes" in str(e):
                            print("\nWe need to authorize the app to make changes to your account.")
                            core.deauthorize(args)
                            mastodon = core.readwrite(args)
                            # retry
                            mastodon.notifications_dismiss(notification["id"])
                            dismissed += 1
                        elif "not found" in str(e):
                            pass
                        elif "Name or service not known" in str(e):
                            error = "Error: the instance name is either misspelled or offline"
                        else:
                            print(e, file=sys.stderr)
                else:
                    print("Dismiss "
                          # + str(notification["id"])
                          + " " + notification["created_at"].strftime("%Y-%m-%d")
                          + " " + notification["account"]["acct"]
                          + " " + notification["type"])

            notifications = mastodon.fetch_next(notifications)

        if error:
            print(error, file=sys.stderr)

        print('Dismissed %d of %d notifications' % (n, total))



#!/usr/bin/env python3
# Copyright (C) 2017-2018  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017       Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from . import core

def login(args):
    """
    Just login to your Mastodon account
    """

    mastodon = core.login(args)

    print("Get user info")

    try:
        user = mastodon.account_verify_credentials()
    except Exception as e:
        print(e, file=sys.stderr)
        if "access token was revoked" in str(e):
            core.deauthorize(args)
            archive(args)
            sys.exit(0)

    print("Login OK")



#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
from progress.bar import Bar
from datetime import timedelta, datetime
from . import core

def mutuals(args):
    """
    List people you're following and who are following you, too.
    """

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required=True, quiet=True, combine=True)

    # Print both error messages if the data is missing
    if "following" not in data or len(data["following"]) == 0:
        print("You need to run 'mastodon-archive archive --with-following'",
              file=sys.stderr)
        sys.exit(7)

    mastodon = core.login(args)

    print("Get user info")

    try:
        user = mastodon.account_verify_credentials()
    except Exception as e:
        if "access token was revoked" in str(e):
            core.deauthorize(args)
             # retry and exit without an error
            archive(args)
            sys.exit(0)
        elif "Name or service not known" in str(e):
            print("Error: the instance name is either misspelled or offline",
              file=sys.stderr)
        else:
            print(e, file=sys.stderr)
        # exit in either case
        sys.exit(1)

    ids = [x["id"] for x in data["following"]]
    lookup = {x["id"]: "%s <%s>" % (x["display_name"] or x["username"],
                                      x["acct"])
              for x in data["following"]}
    relations = mastodon.account_relationships(ids)
    for relation in relations:
        if relation["followed_by"]:
            print(lookup[relation["id"]])



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='mastodon_archive',
    version='1.4.0',
    description="Utility for backing up your Mastodon content",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Alex Schroeder",
    author_email="alex@gnu.org",
    url='https://github.com/kensanata/mastodon-backup#mastodon-archive',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications',
        'Intended Audience :: End Users/Desktop',
        'Environment :: Console',
        'Development Status :: 5 - Production/Stable',
    ],
    packages=["mastodon_archive"],
    entry_points={
        "console_scripts": ["mastodon-archive=mastodon_archive:main"]
    },
    install_requires=[
        "mastodon.py",
        "progress",
        "html2text",
    ],
)



#!/usr/bin/env python3
# Copyright (C) 2018  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import html2text
import re
from . import core

class Found(Exception): pass

def context(args):
    """
    Show the context of a toot: ancestors, the toot itself, and descendants
    """

    url = args.url

    (username, domain) = core.parse(args.user)

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required = True, quiet = True)

    found = None;
    index = {} # mapping ids to statuses
    children = {} # mapping ids to list of children ids

    for collection in ["statuses",
                       "favourites",
                       "mentions"]:
        statuses = data[collection];
        print("Indexing %d %s..." % (len(statuses), collection))
        for status in statuses:

            if status["reblog"] is not None:
                status = status["reblog"]

            # only accept one status per id
            if status["id"] in index:
                pass
                # print("Warning: duplicate id %s" % status["id"], file=sys.stderr)
            else:
                index[status["id"]] = status
                if "in_reply_to_id" in status:
                    if status["in_reply_to_id"] not in children:
                        children[status["in_reply_to_id"]] = [status["id"]]
                    else:
                        children[status["in_reply_to_id"]].append(status["id"])

            for u in [status["uri"],
                      status["url"]]:
                if u == url:
                    found = status
                    # don't break, we want to index them all

    if not found:
        print("The URL/URI was not found", file=sys.stderr)
        sys.exit(5)

    result = [found]

    # add ancestors
    id = found["in_reply_to_id"]

    while id in index:
        status = index[id];
        result.insert(0, status)
        id = status["in_reply_to_id"]

    # add descendants
    try:
        ids = children[found["id"]]
        print(ids)
        while ids:
            id = ids.pop(0)
            result.append(index[id])
            if id in children:
                ids.extend(children[id])
    except:
        if not result:
            print("The status at the provided URL/URI had no context in your archive", file=sys.stderr)
            sys.exit(5)

    for status in result:
        str = '';
        if status["reblog"] is not None:
            str += (status["account"]["display_name"] + "boosted\n")
            status = status["reblog"]
        str += ("%s @%s %s\n" % (
            status["account"]["display_name"],
            status["account"]["username"],
            status["created_at"]))
        str += status["url"] + "\n"
        str += html2text.html2text(status["content"])
        # This forces UTF-8 independent of terminal capabilities, thus
        # avoiding problems with LC_CTYPE=C and other such issues.
        # This works well when redirecting output to a file, which
        # will then be an UTF-8 encoded file.
        sys.stdout.buffer.write(str.encode('utf-8'))



#!/usr/bin/env python3
# Copyright (C) 2017-2018  Alex Schroeder <alex@gnu.org>
# Copyright (C) 2017  Steve Ivy <steveivy@gmail.com>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import itertools
import sys
import os.path
import html2text
import re
from . import core
from urllib.parse import urlparse

def text(args):
    """
    Convert toots to plain text, optionally filtering them
    """

    collection = args.collection
    reverse = args.reverse
    patterns = args.pattern
    combine = args.combine

    (username, domain) = core.parse(args.user)

    media_dir = domain + '.user.' + username
    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required=True, quiet=True, combine=combine)

    def matches(status):
        if status["reblog"] is not None:
            status = status["reblog"]
        for pattern in patterns:
            found = False
            for s in [status["content"],
                      status["account"]["display_name"],
                      status["account"]["username"],
                      status["created_at"]]:
                if re.search(pattern, s, flags=re.IGNORECASE) is not None:
                    found = True
                    continue
            if not found:
                return False
        return True

    if collection == "all":
        statuses = itertools.chain.from_iterable(
            data[collection] for collection in ["statuses", "favourites", "mentions"]
        )
    else:
        statuses = data[collection]

    if len(patterns) > 0:
        statuses = list(filter(matches, statuses))

    statuses = sorted(statuses, reverse=reverse, key=lambda status: status["created_at"])

    for status in statuses:
        str = '';
        if status["reblog"] is not None:
            str += (status["account"]["display_name"] + "boosted\n")
            status = status["reblog"]
        str += ("%s @%s %s\n" % (
            status["account"]["display_name"],
            status["account"]["username"],
            status["created_at"]))
        str += "🔗 " + status["url"] + "\n"
        str += html2text.html2text(status["content"]).strip() + "\n"
        for attachment in status["media_attachments"]:
            # should we check attachment["preview_url"] as well?
            for url in [attachment["url"]]:
                path = urlparse(url).path
                file_name = media_dir + path
                if os.path.isfile(file_name):
                    str += "🖻 " + file_name + "\n"
                elif url not in str:
                    str += "🔗 " + url + "\n"
        str += "\n"
        # This forces UTF-8 independent of terminal capabilities, thus
        # avoiding problems with LC_CTYPE=C and other such issues.
        # This works well when redirecting output to a file, which
        # will then be an UTF-8 encoded file.
        sys.stdout.buffer.write(str.encode('utf-8'))



#!/usr/bin/env python3
# Copyright (C) 2019  Alex Schroeder <alex@gnu.org>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import sys
import os.path
import math
from datetime import timedelta, datetime
from . import core

def split(args):
    """
    Split older toots into a new file
    """

    confirmed = args.confirmed

    if not confirmed:

        print("This is a dry run and nothing will be moved.\n"
              "Instead, we'll just list what would have happened.\n"
              "Use --confirmed to actually do it.")

    (username, domain) = args.user.split('@')

    status_file = domain + '.user.' + username + '.json'
    data = core.load(status_file, required = True)
    older_data = {}

    n = 0
    older_status_file = ''
    while True:
        older_status_file = domain + '.user.' + username + '.' + str(n) + '.json'
        if os.path.exists(older_status_file):
            n = n + 1
        else:
            break

    delta = timedelta(weeks = args.weeks)
    cutoff = datetime.today() - delta

    print("Older than " + str(cutoff))

    n_statuses = 0

    for collection in ["statuses",
                       "favourites",
                       "mentions"]:
        statuses = list(filter(
            lambda status: datetime.strptime(status["created_at"][0:10], "%Y-%m-%d") >= cutoff,
            data[collection]))
        older_statuses = list(filter(
            lambda status: datetime.strptime(status["created_at"][0:10], "%Y-%m-%d") < cutoff,
            data[collection]))

        data[collection] = statuses
        older_data[collection] = older_statuses

        moved = len(older_statuses)
        print(collection + ": " + str(moved))
        n_statuses += moved

    if confirmed and n_statuses > 0:

        print("Saving " + status_file)
        core.save(status_file, data)
        print("Saving " + older_status_file)
        core.save(older_status_file, older_data)

    elif confirmed:

        print("No older statuses to move")

    else:

        print("Would have saved this to " + older_status_file)
